@@ -20,23 +20,21 @@ import copy
 import logging
 import os
 import platform
-import shlex
 
 from botocore import __version__
 import botocore.config
 import botocore.credentials
 import botocore.client
-from botocore.endpoint import EndpointCreator
-from botocore.exceptions import EventNotFound, ConfigNotFound, ProfileNotFound
+from botocore.exceptions import ConfigNotFound, ProfileNotFound
 from botocore import handlers
 from botocore.hooks import HierarchicalEmitter, first_non_none_response
-from botocore.loaders import Loader
-from botocore.provider import get_provider
+from botocore.loaders import create_loader
 from botocore.parsers import ResponseParserFactory
 from botocore import regions
 from botocore.model import ServiceModel
-import botocore.service
+from botocore import paginate
 from botocore import waiter
+from botocore import retryhandler, translate
 
 
 class Session(object):
@@ -50,79 +48,59 @@ class Session(object):
     :ivar profile: The current profile.
     """
 
-    AllEvents = {
-        'after-call': '.%s.%s',
-        'after-parsed': '.%s.%s.%s.%s',
-        'before-parameter-build': '.%s.%s',
-        'before-call': '.%s.%s',
-        'service-created': '',
-        'service-data-loaded': '.%s',
-        'creating-endpoint': '.%s',
-        'before-auth': '.%s',
-        'needs-retry': '.%s.%s',
-    }
-    """
-    A dictionary where each key is an event name and the value
-    is the formatting string used to construct a new event.
-    """
+    #: A default dictionary that maps the logical names for session variables
+    #: to the specific environment variables and configuration file names
+    #: that contain the values for these variables.
+    #: When creating a new Session object, you can pass in your own dictionary
+    #: to remap the logical names or to add new logical names.  You can then
+    #: get the current value for these variables by using the
+    #: ``get_config_variable`` method of the :class:`botocore.session.Session`
+    #: class.
+    #: These form the keys of the dictionary.  The values in the dictionary
+    #: are tuples of (<config_name>, <environment variable>, <default value>,
+    #: <conversion func>).
+    #: The conversion func is a function that takes the configuration value
+    #: as an argument and returns the converted value.  If this value is
+    #: None, then the configuration value is returned unmodified.  This
+    #: conversion function can be used to type convert config values to
+    #: values other than the default values of strings.
+    #: The ``profile`` and ``config_file`` variables should always have a
+    #: None value for the first entry in the tuple because it doesn't make
+    #: sense to look inside the config file for the location of the config
+    #: file or for the default profile to use.
+    #: The ``config_name`` is the name to look for in the configuration file,
+    #: the ``env var`` is the OS environment variable (``os.environ``) to
+    #: use, and ``default_value`` is the value to use if no value is otherwise
+    #: found.
+    SESSION_VARIABLES = {
+        # logical:  config_file, env_var,        default_value, conversion_func
+        'profile': (None, ['AWS_DEFAULT_PROFILE', 'AWS_PROFILE'], None, None),
+        'region': ('region', 'AWS_DEFAULT_REGION', None, None),
+        'data_path': ('data_path', 'AWS_DATA_PATH', None, None),
+        'config_file': (None, 'AWS_CONFIG_FILE', '~/.aws/config', None),
+        'ca_bundle': ('ca_bundle', 'AWS_CA_BUNDLE', None, None),
 
-    SessionVariables = {
-        # logical:  config_file, env_var,        default_value
-        'profile': (None, 'BOTO_DEFAULT_PROFILE', None),
-        'region': ('region', 'BOTO_DEFAULT_REGION', None),
-        'data_path': ('data_path', 'BOTO_DATA_PATH', None),
-        'config_file': (None, 'AWS_CONFIG_FILE', '~/.aws/config'),
-        'provider': ('provider', 'BOTO_PROVIDER_NAME', 'aws'),
-
-        # These variables are intended for internal use so don't have any
-        # user settable values.
         # This is the shared credentials file amongst sdks.
-        'credentials_file': (None, None, '~/.aws/credentials'),
+        'credentials_file': (None, 'AWS_SHARED_CREDENTIALS_FILE',
+                             '~/.aws/credentials', None),
 
         # These variables only exist in the config file.
 
         # This is the number of seconds until we time out a request to
         # the instance metadata service.
-        'metadata_service_timeout': ('metadata_service_timeout', None, 1),
+        'metadata_service_timeout': ('metadata_service_timeout',
+                                     'AWS_METADATA_SERVICE_TIMEOUT', 1, int),
         # This is the number of request attempts we make until we give
         # up trying to retrieve data from the instance metadata service.
         'metadata_service_num_attempts': ('metadata_service_num_attempts',
-                                          None, 1),
-        }
-    """
-    A default dictionary that maps the logical names for session variables
-    to the specific environment variables and configuration file names
-    that contain the values for these variables.
+                                          'AWS_METADATA_SERVICE_NUM_ATTEMPTS', 1, int),
+    }
 
-    When creating a new Session object, you can pass in your own dictionary to
-    remap the logical names or to add new logical names.  You can then get the
-    current value for these variables by using the ``get_config_variable``
-    method of the :class:`botocore.session.Session` class.
-    The default set of logical variable names are:
-
-    * profile - Default profile name you want to use.
-    * region - Default region name to use, if not otherwise specified.
-    * data_path - Additional directories to search for data files.
-    * config_file - Location of a Boto config file.
-    * provider - The name of the service provider (e.g. aws)
-
-    These form the keys of the dictionary.  The values in the dictionary
-    are tuples of (<config_name>, <environment variable>, <default value).
-    The ``profile`` and ``config_file`` variables should always have a
-    None value for the first entry in the tuple because it doesn't make
-    sense to look inside the config file for the location of the config
-    file or for the default profile to use.
-
-    The ``config_name`` is the name to look for in the configuration file,
-    the ``env var`` is the OS environment variable (``os.environ``) to
-    use, and ``default_value`` is the value to use if no value is otherwise
-    found.
-    """
-
-    FmtString = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
+    #: The default format string to use when configuring the botocore logger.
+    LOG_FORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
 
     def __init__(self, session_vars=None, event_hooks=None,
-                 include_builtin_handlers=True, loader=None):
+                 include_builtin_handlers=True, profile=None):
         """
         Create a new Session object.
 
@@ -130,7 +108,7 @@ class Session(object):
         :param session_vars: A dictionary that is used to override some or all
             of the environment variables associated with this session.  The
             key/value pairs defined in this dictionary will override the
-            corresponding variables defined in ``SessionVariables``.
+            corresponding variables defined in ``SESSION_VARIABLES``.
 
         :type event_hooks: BaseEventHooks
         :param event_hooks: The event hooks object to use. If one is not
@@ -140,8 +118,14 @@ class Session(object):
         :type include_builtin_handlers: bool
         :param include_builtin_handlers: Indicates whether or not to
             automatically register builtin handlers.
+
+        :type profile: str
+        :param profile: The name of the profile to use for this
+            session.  Note that the profile can only be set when
+            the session is created.
+
         """
-        self.session_var_map = copy.copy(self.SessionVariables)
+        self.session_var_map = copy.copy(self.SESSION_VARIABLES)
         if session_vars:
             self.session_var_map.update(session_vars)
         if event_hooks is None:
@@ -153,21 +137,19 @@ class Session(object):
         self.user_agent_name = 'Botocore'
         self.user_agent_version = __version__
         self.user_agent_extra = ''
+        # The _profile attribute is just used to cache the value
+        # of the current profile to avoid going through the normal
+        # config lookup process each access time.
         self._profile = None
         self._config = None
         self._credentials = None
         self._profile_map = None
-        self._provider = None
         # This is a dict that stores per session specific config variable
         # overrides via set_config_variable().
         self._session_instance_vars = {}
-        if loader is None:
-            loader = Loader()
-        self._loader = loader
-        # _data_paths_added is used to track whether or not we added
-        # extra paths to the loader.  We will do this lazily
-        # only when we ask for the loader.
-        self._data_paths_added = False
+        if profile is not None:
+            self._session_instance_vars['profile'] = profile
+        self._client_config = None
         self._components = ComponentLocator()
         self._register_components()
 
@@ -189,20 +171,17 @@ class Session(object):
     def _register_data_loader(self):
         self._components.lazy_register_component(
             'data_loader',
-            lambda:  Loader(self.get_config_variable('data_path') or ''))
+            lambda:  create_loader(self.get_config_variable('data_path')))
 
     def _register_endpoint_resolver(self):
         self._components.lazy_register_component(
             'endpoint_resolver',
-            lambda:  regions.EndpointResolver(self.get_data('aws/_endpoints')))
+            lambda:  regions.EndpointResolver(self.get_data('_endpoints')))
 
     def _register_response_parser_factory(self):
         self._components.register_component('response_parser_factory',
                                             ResponseParserFactory())
 
-    def _reset_components(self):
-        self._register_components()
-
     def _register_builtin_handlers(self, events):
         for spec in handlers.BUILTIN_HANDLERS:
             if len(spec) == 2:
@@ -212,16 +191,9 @@ class Session(object):
                 event_name, handler, register_type = spec
                 if register_type is handlers.REGISTER_FIRST:
                     self._events.register_first(event_name, handler)
-                elif register_first is handlers.REGISTER_LAST:
+                elif register_type is handlers.REGISTER_LAST:
                     self._events.register_last(event_name, handler)
 
-    @property
-    def provider(self):
-        if self._provider is None:
-            self._provider = get_provider(
-                self, self.get_config_variable('provider'))
-        return self._provider
-
     @property
     def available_profiles(self):
         return list(self._build_profile_map().keys())
@@ -236,20 +208,13 @@ class Session(object):
 
     @property
     def profile(self):
+        if self._profile is None:
+            profile = self.get_config_variable('profile')
+            self._profile = profile
         return self._profile
 
-    @profile.setter
-    def profile(self, profile):
-        # Since provider can be specified in profile, changing the
-        # profile should reset the provider.
-        self._provider = None
-        self._profile = profile
-        # Need to potentially reload the config file/creds.
-        self._reset_components()
-
     def get_config_variable(self, logical_name,
-                            methods=('instance', 'env', 'config'),
-                            default=None):
+                            methods=('instance', 'env', 'config')):
         """
         Retrieve the value associated with the specified logical_name
         from the environment or the config file.  Values found in the
@@ -269,48 +234,54 @@ class Session(object):
             by supplying a different value to this parameter.
             Valid choices are: instance|env|config
 
-        :param default: The default value to return if there is no
-            value associated with the config file.  This value will
-            override any default value specified in ``SessionVariables``.
-
-        :returns: str value of variable of None if not defined.
+        :returns: value of variable or None if not defined.
 
         """
+        # Handle all the short circuit special cases first.
+        if logical_name not in self.session_var_map:
+            return
+        # Do the actual lookups.  We need to handle
+        # 'instance', 'env', and 'config' locations, in that order.
         value = None
-        # There's two types of defaults here.  One if the
-        # default value specified in the SessionVariables.
-        # The second is an explicit default value passed into this
-        # function (the default parameter).
-        # config_default is tracking the default value specified
-        # in the SessionVariables.
-        config_default = None
-        if logical_name in self.session_var_map:
-            # Short circuit case, check if the var has been explicitly
-            # overriden via set_config_variable.
-            if 'instance' in methods and \
-                    logical_name in self._session_instance_vars:
-                return self._session_instance_vars[logical_name]
-            config_name, envvar_name, config_default = self.session_var_map[
-                logical_name]
-            if logical_name in ('config_file', 'profile'):
-                config_name = None
-            if logical_name == 'profile' and self._profile:
-                value = self._profile
-            elif 'env' in methods and envvar_name and envvar_name in os.environ:
-                value = os.environ[envvar_name]
-            elif 'config' in methods:
-                if config_name:
-                    config = self.get_scoped_config()
-                    value = config.get(config_name)
-        # If we don't have a value at this point, we need to try to assign
-        # a default value.  An explicit default argument will win over the
-        # default value from SessionVariables.
-        if value is None and default is not None:
-            value = default
-        if value is None and config_default is not None:
-            value = config_default
+        var_config = self.session_var_map[logical_name]
+        if self._found_in_instance_vars(methods, logical_name):
+            return self._session_instance_vars[logical_name]
+        elif self._found_in_env(methods, var_config):
+            value = self._retrieve_from_env(var_config[1], os.environ)
+        elif self._found_in_config_file(methods, var_config):
+            value = self.get_scoped_config()[var_config[0]]
+        if value is None:
+            value = var_config[2]
+        if var_config[3] is not None:
+            value = var_config[3](value)
         return value
 
+    def _found_in_instance_vars(self, methods, logical_name):
+        if 'instance' in methods:
+            return logical_name in self._session_instance_vars
+        return False
+
+    def _found_in_env(self, methods, var_config):
+        return (
+            'env' in methods and
+            var_config[1] is not None and
+            self._retrieve_from_env(var_config[1], os.environ) is not None)
+
+    def _found_in_config_file(self, methods, var_config):
+        if 'config' in methods and var_config[0] is not None:
+            return var_config[0] in self.get_scoped_config()
+        return False
+
+    def _retrieve_from_env(self, names, environ):
+        # We need to handle the case where names is either
+        # a single value or a list of variables.
+        if not isinstance(names, list):
+            names = [names]
+        for name in names:
+            if name in environ:
+                return environ[name]
+        return None
+
     def set_config_variable(self, logical_name, value):
         """Set a configuration variable to a specific value.
 
@@ -331,13 +302,12 @@ class Session(object):
 
         :type logical_name: str
         :param logical_name: The logical name of the session variable
-            you want to set.  These are the keys in ``SessionVariables``.
+            you want to set.  These are the keys in ``SESSION_VARIABLES``.
         :param value: The value to associate with the config variable.
 
         """
         self._session_instance_vars[logical_name] = value
 
-
     def get_scoped_config(self):
         """
         Returns the config values from the config file scoped to the current
@@ -409,6 +379,26 @@ class Session(object):
                 pass
         return self._config
 
+    def get_default_client_config(self):
+        """Retrieves the default config for creating clients
+
+        :rtype: botocore.client.Config
+        :returns: The default client config object when creating clients. If
+            the value is ``None`` then there is no default config object
+            attached to the session.
+        """
+        return self._client_config
+
+    def set_default_client_config(self, client_config):
+        """Sets the default config for creating clients
+
+        :type client_config: botocore.client.Config
+        :param client_config: The default client config object when creating
+            clients. If the value is ``None`` then there is no default config
+            object attached to the session.
+        """
+        self._client_config = client_config
+
     def set_credentials(self, access_key, secret_key, token=None):
         """
         Manually create credentials for this session.  If you would
@@ -499,27 +489,32 @@ class Session(object):
 
         """
         service_description = self.get_service_data(service_name, api_version)
-        return ServiceModel(service_description)
+        return ServiceModel(service_description, service_name=service_name)
 
     def get_waiter_model(self, service_name, api_version=None):
         loader = self.get_component('data_loader')
-        latest = loader.determine_latest('%s/%s' % (
-            self.provider.name, service_name), api_version)
-        waiter_path = latest.replace('.normal', '.waiters')
-        waiter_config = loader.load_data(waiter_path)
+        waiter_config = loader.load_service_model(
+            service_name, 'waiters-2', api_version)
         return waiter.WaiterModel(waiter_config)
 
+    def get_paginator_model(self, service_name, api_version=None):
+        loader = self.get_component('data_loader')
+        paginator_config = loader.load_service_model(
+            service_name, 'paginators-1', api_version)
+        return paginate.PaginatorModel(paginator_config)
+
     def get_service_data(self, service_name, api_version=None):
         """
         Retrieve the fully merged data associated with a service.
         """
-        data_path = '%s/%s' % (self.provider.name, service_name)
+        data_path = service_name
         service_data = self.get_component('data_loader').load_service_model(
             data_path,
+            type_name='service-2',
             api_version=api_version
         )
-        event_name = self.create_event('service-data-loaded', service_name)
-        self._events.emit(event_name, service_data=service_data,
+        self._events.emit('service-data-loaded.%s' % service_name,
+                          service_data=service_data,
                           service_name=service_name, session=self)
         return service_data
 
@@ -527,25 +522,8 @@ class Session(object):
         """
         Return a list of names of available services.
         """
-        data_path = '%s' % self.provider.name
         return self.get_component('data_loader')\
-                .list_available_services(data_path)
-
-    def get_service(self, service_name, api_version=None):
-        """
-        Get information about a service.
-
-        :type service_name: str
-        :param service_name: The name of the service (e.g. 'ec2')
-
-        :returns: :class:`botocore.service.Service`
-        """
-        service = botocore.service.get_service(self, service_name,
-                                               self.provider,
-                                               api_version=api_version)
-        event = self.create_event('service-created')
-        self._events.emit(event, service=service)
-        return service
+            .list_available_services(type_name='service-2')
 
     def set_debug_logger(self, logger_name='botocore'):
         """
@@ -574,7 +552,7 @@ class Session(object):
         :type format_string: str
         :param format_string: The format string to use for the log
             formatter.  If none is provided this will default to
-            ``self.FmtString``.
+            ``self.LOG_FORMAT``.
 
         """
         log = logging.getLogger(logger_name)
@@ -585,7 +563,7 @@ class Session(object):
 
         # create formatter
         if format_string is None:
-            format_string = self.FmtString
+            format_string = self.LOG_FORMAT
         formatter = logging.Formatter(format_string)
 
         # add formatter to ch
@@ -614,7 +592,7 @@ class Session(object):
         ch.setLevel(log_level)
 
         # create formatter
-        formatter = logging.Formatter(self.FmtString)
+        formatter = logging.Formatter(self.LOG_FORMAT)
 
         # add formatter to ch
         ch.setFormatter(formatter)
@@ -693,44 +671,6 @@ class Session(object):
                                 unique_id=unique_id,
                                 unique_id_uses_count=unique_id_uses_count)
 
-    def register_event(self, event_name, fmtstr):
-        """
-        Register a new event.  The event will be added to ``AllEvents``
-        and will then be able to be created using ``create_event``.
-
-        :type event_name: str
-        :param event_name: The base name of the event.
-
-        :type fmtstr: str
-        :param fmtstr: The formatting string for the event.
-        """
-        if event_name not in self.AllEvents:
-            self.AllEvents[event_name] = fmtstr
-
-    def create_event(self, event_name, *fmtargs):
-        """
-        Creates a new event string that can then be emitted.
-        You could just create it manually, since it's just
-        a string but this helps to define the range of known events.
-
-        :type event_name: str
-        :param event_name: The base name of the new event.
-
-        :type fmtargs: tuple
-        :param fmtargs: A tuple of values that will be used as the
-            arguments pass to the string formatting operation.  The
-            actual values passed depend on the type of event you
-            are creating.
-        """
-        if event_name in self.AllEvents:
-            fmt_string = self.AllEvents[event_name]
-            if fmt_string:
-                event = event_name + (fmt_string % fmtargs)
-            else:
-                event = event_name
-            return event
-        raise EventNotFound(event_name=event_name)
-
     def emit(self, event_name, **kwargs):
         return self._events.emit(event_name, **kwargs)
 
@@ -750,7 +690,7 @@ class Session(object):
     def create_client(self, service_name, region_name=None, api_version=None,
                       use_ssl=True, verify=None, endpoint_url=None,
                       aws_access_key_id=None, aws_secret_access_key=None,
-                      aws_session_token=None):
+                      aws_session_token=None, config=None):
         """Create a botocore client.
 
         :type service_name: string
@@ -769,26 +709,28 @@ class Session(object):
             of the client.
 
         :type use_ssl: boolean
-        :param use_ssl: Whether or not to use SSL.  By default, SSL is used.  Note that
-            not all services support non-ssl connections.
+        :param use_ssl: Whether or not to use SSL.  By default, SSL is used.
+            Note that not all services support non-ssl connections.
 
         :type verify: boolean/string
-        :param verify: Whether or not to verify SSL certificates.  By default SSL certificates
-            are verified.  You can provide the following values:
+        :param verify: Whether or not to verify SSL certificates.
+            By default SSL certificates are verified.  You can provide the
+            following values:
 
             * False - do not validate SSL certificates.  SSL will still be
               used (unless use_ssl is False), but SSL certificates
               will not be verified.
             * path/to/cert/bundle.pem - A filename of the CA cert bundle to
-              uses.  You can specify this argument if you want to use a different
-              CA cert bundle than the one used by botocore.
+              uses.  You can specify this argument if you want to use a
+              different CA cert bundle than the one used by botocore.
 
         :type endpoint_url: string
-        :param endpoint_url: The complete URL to use for the constructed client.
-            Normally, botocore will automatically construct the appropriate URL
-            to use when communicating with a service.  You can specify a
-            complete URL (including the "http/https" scheme) to override this
-            behavior.  If this value is provided, then ``use_ssl`` is ignored.
+        :param endpoint_url: The complete URL to use for the constructed
+            client.  Normally, botocore will automatically construct the
+            appropriate URL to use when communicating with a service.  You can
+            specify a complete URL (including the "http/https" scheme) to
+            override this behavior.  If this value is provided, then
+            ``use_ssl`` is ignored.
 
         :type aws_access_key_id: string
         :param aws_access_key_id: The access key to use when creating
@@ -805,40 +747,54 @@ class Session(object):
         :param aws_session_token: The session token to use when creating
             the client.  Same semantics as aws_access_key_id above.
 
+        :type config: botocore.client.Config
+        :param config: Advanced client configuration options. If a value
+            is specified in the client config, its value will take precedence
+            over environment variables and configuration values, but not over
+            a value passed explicitly to the method. If a default config
+            object is set on the session, the config object used when creating
+            the client will be the result of calling ``merge()`` on the
+            default config with the config provided to this call.
+
         :rtype: botocore.client.BaseClient
         :return: A botocore client instance
 
         """
+        default_client_config = self.get_default_client_config()
+        # If a config is provided and a default config is set, then
+        # use the config resulting from merging the two.
+        if config is not None and default_client_config is not None:
+            config = default_client_config.merge(config)
+        # If a config was not provided then use the default
+        # client config from the session
+        elif default_client_config is not None:
+            config = default_client_config
+
+        if region_name is None:
+            if config and config.region_name is not None:
+                region_name = config.region_name
+            else:
+                region_name = self.get_config_variable('region')
         loader = self.get_component('data_loader')
-        endpoint_creator = self._create_endpoint_creator(aws_access_key_id,
-                                                         aws_secret_access_key,
-                                                         aws_session_token)
         event_emitter = self.get_component('event_emitter')
         response_parser_factory = self.get_component(
             'response_parser_factory')
-        client_creator = botocore.client.ClientCreator(loader, endpoint_creator,
-                                                       event_emitter,
-                                                       response_parser_factory)
-        client = client_creator.create_client(service_name, region_name, use_ssl,
-                                              endpoint_url, verify,
-                                              aws_access_key_id,
-                                              aws_secret_access_key,
-                                              aws_session_token)
-        return client
-
-    def _create_endpoint_creator(self, aws_access_key_id, aws_secret_access_key,
-                                 aws_session_token):
-        resolver = self.get_component('endpoint_resolver')
-        region = self.get_config_variable('region')
-        event_emitter = self.get_component('event_emitter')
-        if aws_secret_access_key is None:
-            credentials = self.get_credentials()
+        if aws_secret_access_key is not None:
+            credentials = botocore.credentials.Credentials(
+                access_key=aws_access_key_id,
+                secret_key=aws_secret_access_key,
+                token=aws_session_token)
         else:
-            credentials = None
-        user_agent= self.user_agent()
-        endpoint_creator = EndpointCreator(resolver, region, event_emitter,
-                                           credentials, user_agent)
-        return endpoint_creator
+            credentials = self.get_credentials()
+        endpoint_resolver = self.get_component('endpoint_resolver')
+        client_creator = botocore.client.ClientCreator(
+            loader, endpoint_resolver, self.user_agent(), event_emitter,
+            retryhandler, translate, response_parser_factory)
+        client = client_creator.create_client(
+            service_name, region_name, use_ssl, endpoint_url, verify,
+            credentials, scoped_config=self.get_scoped_config(),
+            client_config=config, api_version=api_version)
+        return client
 
 
 class ComponentLocator(object):
@@ -849,8 +805,12 @@ class ComponentLocator(object):
 
     def get_component(self, name):
         if name in self._deferred:
-            factory = self._deferred.pop(name)
+            factory = self._deferred[name]
             self._components[name] = factory()
+            # Only delete the component from the deferred dict after
+            # successfully creating the object from the factory as well as
+            # injecting the instantiated value into the _components dict.
+            del self._deferred[name]
         try:
             return self._components[name]
         except KeyError:
